# vim:set et sw=4 ts=4:
# SPDX-FileCopyrightText: 2024-present Atri Bhattacharya <atrib@duck.com>
#
# SPDX-License-Identifier: MIT
#

"""
Helper function to compile latex file in a temp dir
"""

import logging as log
import re
import shutil
import subprocess
import sys
from pathlib import Path

from rich.live import Live
from rich.spinner import Spinner

# Match only lines beginning with INPUT
INPUT_RE = re.compile(r"^INPUT")
INPUT_STY = re.compile(r"^INPUT\s.*.(cls|def|sty)")
INPUT_FONTS = re.compile(r"^INPUT\s.*.(pfb|tfm)")
FONT_PUBLIC = re.compile(r"/public/.*/")


def run_latexmk(filename, mode, compdir):
    """Helper function to actually compile the latex file in a tmpdir"""
    # Generate fls file from tex file by running latexmk
    latexmk_cmd = [
        shutil.which("latexmk"),
        f"-{mode}",
        "-f",
        "-cd",
        f"-outdir={compdir}",
        "-interaction=nonstopmode",
        filename.name,
    ]
    try:
        with Live(
            Spinner("dots2", text="Compiling LaTeX project"),
            transient=True,
        ):
            subprocess.run(
                latexmk_cmd,
                capture_output=True,
                encoding="utf-8",
                check=True,
            )
    except OSError as err:
        log.critical("%s", err.strerror)
        sys.exit(1)
    except subprocess.CalledProcessError as err:
        log.critical(
            "Error: %s failed with the following output:\n%s",
            err.cmd[0],
            err.stdout,
        )
        sys.exit(1)
    except TypeError as err:  # Typically when latexmk is missing and shutil
                              # gets a None as the first elem of cmdline list
        log.critical("%s", err)
        log.critical("Is latexmk installed and in PATH?")
        sys.exit(1)

    log.info(
        "LaTeX project successfully compiled with: %s", " ".join(latexmk_cmd)
    )
    fls_path = Path(compdir) / f"{filename.stem}.fls"
    log.debug("%s generated", fls_path.as_posix())
    return fls_path


def fls_input_files(fls_fileobj, lof_excl, skip_files, *, sty_files=False):
    """Helper function to return list on files marked as 'INPUT' in fls file"""
    deps = set()
    pkgs = {"System": set(), "Local": set()}
    for line in fls_fileobj:
        if INPUT_RE.match(line):
            p = Path(line.split()[-1])
            if (
                not p.is_absolute()
                and (p.as_posix() not in deps)
                and (p.as_posix() not in lof_excl)
                and (p.suffix not in skip_files)
            ):
                deps.add(p.as_posix())
                log.info("Add file: %s", p.as_posix())

        if sty_files:
            if INPUT_STY.match(line):
                p = Path(line.split()[-1])
                if p.is_absolute():
                    # Base is not a (La)TeX package; it is installed with even
                    # the most basic TeXlive/MikTeX installation
                    if (pdir := p.parent.name) != "base":
                        pkgs["System"].add(pdir)
                else:
                    pkgs["Local"].add(p.stem)
            elif INPUT_FONTS.match(line):
                p = Path(line.split()[-1])
                if p.is_absolute():
                    try:
                        fontdir = (
                            FONT_PUBLIC.search(str(p)).group(0).split("/")[2]
                        )
                    except AttributeError:
                        fontdir = p.parent.name
                    pkgs["System"].add(fontdir)

    return list(deps), pkgs



# vim: set ai et ts=4 sw=4 tw=80:
# SPDX-FileCopyrightText: 2024-present Atri Bhattacharya <atrib@duck.com>
#
# SPDX-License-Identifier: MIT

"""Common fixtures"""

import os
import shutil

import pytest

from tartex.tartex import TarTeX


@pytest.fixture
def sample_texfile():
    """Pytest fixture: TarTeX with just a tex file for parameter"""
    return TarTeX(["some_file.tex"])


# Data copying to tmpdir to allow using with pytest
# Note that relative paths to the data do not work as it is never certain which
# location pytest is run from
# Based on https://stackoverflow.com/a/29631801
@pytest.fixture
def datadir(tmpdir, request):
    """
    Fixture responsible for searching a folder with the same name of test
    module and, if available, moving all contents to a temporary directory so
    tests can use them freely.
    """
    filename = request.module.__file__
    test_dir, _ = os.path.splitext(filename)

    if os.path.isdir(test_dir):
        shutil.copytree(test_dir, tmpdir, dirs_exist_ok=True)

    return tmpdir



# vim:set et sw=4 ts=4:
"""
Module to help users with completion syntax for tartex
"""

import contextlib
import os
from pathlib import Path
from shutil import copy2

from rich import print as richprint

from tartex.__about__ import __appname__ as APPNAME  # noqa

COMPFILE = {
    "bash": Path(f"bash-completion/completions/{APPNAME}"),
    "fish": Path(f"fish/vendor_completions.d/{APPNAME}.fish"),
    "zsh": Path(f"zsh-completions/_{APPNAME}"),
}


class Completion:

    """Methods for helping users print or install shell completion"""

    def __init__(self, shell_name, filename):
        """Initialise"""
        self.shell = shell_name
        self.completion_file = Path(__file__).parent.joinpath("data", filename)
        self.data = self.completion_file.read_text(encoding="utf-8")

        install_root = Path(
            os.getenv("XDG_DATA_HOME")
            or Path.home().joinpath(".local", "share")
        )
        self.install_dir = install_root.joinpath(COMPFILE[self.shell]).parent
        self.install_filename = COMPFILE[self.shell].name

    def install(self, install_dir=None):
        """Install completion to path"""
        path = Path(install_dir or self.install_dir)
        with contextlib.suppress(FileExistsError):
            os.makedirs(path)
        inst_path = Path(
            copy2(self.completion_file, path.joinpath(self.install_filename))
        )
        richprint(
            f"✓ Completion file for [bold]{self.shell}[/] shell installed to"
            f" [link={inst_path.parent.as_uri()}]{inst_path.parent}[/]"
        )


class BashCompletion(Completion):

    """Completion for bash"""

    def __init__(self):
        """Initialise"""
        Completion.__init__(
            self, shell_name="bash", filename="tartex-completion.bash"
        )


class ZshCompletion(Completion):

    """Completion for zsh shell"""

    def __init__(self):
        """Initialise"""
        Completion.__init__(
            self, shell_name="zsh", filename="tartex-completion.zsh"
        )


class FishCompletion(Completion):

    """Completion for fish shell"""

    def __init__(self):
        """Initialise"""
        Completion.__init__(
            self, shell_name="fish", filename="tartex-completion.fish"
        )



# vim: set ai et ts=4 sw=4 tw=80:
# SPDX-FileCopyrightText: 2024-present Atri Bhattacharya <atrib@duck.com>
#
# SPDX-License-Identifier: MIT
#
"""Tests for tarball generation from basic latex files"""
import os
import tarfile as tar
from pathlib import Path

import pytest

from tartex.tartex import TAR_DEFAULT_COMP, TarTeX


@pytest.fixture
def default_target(datadir):
    return Path(datadir) / "test"


@pytest.fixture
def default_tartex_obj(datadir, default_target):
    return TarTeX(
        [
            (Path(datadir) / "basic_latex.tex").as_posix(),
            "-v",
            "-s",
            "-o",
            default_target.as_posix(),
        ]
    )


class TestBasicLaTeX:
    """Tests checking tar file generation from a basic latex file"""

    def test_gen_tar(self, default_target, default_tartex_obj):
        """Should include a single file in tarball"""
        output = default_target.with_suffix(".tar.gz")
        t = default_tartex_obj
        t.tar_files()
        assert output.exists() is True
        with tar.open(output) as rat:
            assert len(rat.getnames()) == 1

    def test_diff_target_dir(self, tmpdir, datadir):
        # Make a new dir inside tmpdir
        destdir = tmpdir / "dest"
        os.mkdir(destdir)
        t = TarTeX(
            [
                (Path(datadir) / "basic_latex.tex").as_posix(),
                "-v",
                "-s",
                "-o",
                str(destdir / "output.tar.gz"),
            ]
        )
        t.tar_files()
        assert t.tar_file.with_suffix(f".tar.{t.tar_ext}").exists()

    def test_verbose_debug(self, datadir):
        """
        Test verbose value set by -vv option
        """
        t = TarTeX([(Path(datadir) / "basic_latex.tex").as_posix(), "-vv"])
        assert t.args.verbose == 2


# These tests involve repeatedly compiling LaTeX files, thus can be slow
@pytest.mark.slow
class TestTarConflict:
    """Tests checking resolutions for tar file name conflicts"""

    def test_sol_default(self, default_tartex_obj, capsys, monkeypatch):
        """Test default (empty) user response"""
        t_con = default_tartex_obj
        t_con.tar_files()

        # Monkeypatch empty response for input
        monkeypatch.setattr("rich.prompt.Prompt.ask", lambda _: "")

        # Trying to create tar file again will lead to conflic res dialog
        # Blank user input (from monkeypatch) will raise SystemExit
        with pytest.raises(SystemExit) as exc:
            t_con.tar_files()

        assert "Not overwriting existing tar file" in capsys.readouterr().err
        assert exc.value.code == 1

    def test_sol_quit(self, default_tartex_obj, capsys, monkeypatch):
        """Test when user response is 'q'"""
        t_con = default_tartex_obj
        t_con.tar_files()

        # Monkeypatch empty response for input
        monkeypatch.setattr("rich.prompt.Prompt.ask", lambda _: "q")

        # Trying to create tar file again will lead to conflic res dialog
        # Blank user input (from monkeypatch) will raise SystemExit
        with pytest.raises(SystemExit) as exc:
            t_con.tar_files()

        assert "Not overwriting existing tar file" in capsys.readouterr().err
        assert exc.value.code == 1

    def test_sol_overwrite(self, default_tartex_obj, monkeypatch):
        """Test overwrite resolution"""
        t_con = default_tartex_obj
        t_con.tar_files()

        # Monkeypatch empty response for input
        monkeypatch.setattr("rich.prompt.Prompt.ask", lambda _: "o")
        t_con.tar_files()
        output = t_con.tar_file.with_suffix(f".tar.{TAR_DEFAULT_COMP}")
        assert output.exists() is True
        with tar.open(output) as rat:
            assert len(rat.getnames()) == 1

    def test_sol_newname_ok(self, default_tartex_obj, tmpdir, monkeypatch):
        """Test entering new name that works"""
        t_con = default_tartex_obj
        t_con.tar_files()

        output = str(tmpdir / "new.tar.gz")
        # Monkeypatch responses for choosing a new file name
        user_inputs = iter(["c", output])
        monkeypatch.setattr(
            "rich.prompt.Prompt.ask", lambda _: next(user_inputs)
        )
        t_con.tar_files()
        assert Path(output).exists() is True
        with tar.open(output) as rat:
            assert len(rat.getnames()) == 1

    def test_sol_newname_old(
        self, default_tartex_obj, tmpdir, capsys, monkeypatch
    ):
        """Test error when entering new name that is same as the old name"""
        t_con = default_tartex_obj
        t_con.tar_files()

        output = str(tmpdir / "test.tar.gz")
        assert Path(output).exists() is True

        # Monkeypatch responses for choosing file name same as original
        user_inputs = iter(["c", output])
        monkeypatch.setattr(
            "rich.prompt.Prompt.ask", lambda _: next(user_inputs)
        )
        with pytest.raises(SystemExit) as exc:
            t_con.tar_files()

        assert "New name entered is also the same" in capsys.readouterr().err
        assert exc.value.code == 1

    def test_sol_newext(self, default_tartex_obj, tmpdir, monkeypatch):
        """Test new name with just the file ext changed"""
        t_con = default_tartex_obj
        t_con.tar_files()

        output = str(tmpdir / "test.tar.gz")
        assert Path(output).exists() is True

        output = output.replace(".gz", ".xz")
        # Monkeypatch responses for choosing file name same as original
        user_inputs = iter(["c", output])
        monkeypatch.setattr(
            "rich.prompt.Prompt.ask", lambda _: next(user_inputs)
        )
        t_con.tar_files()
        assert Path(output).exists() is True
        assert t_con.tar_ext == "xz"
        with tar.open(output) as rat:
            assert len(rat.getnames()) == 1



# vim:set et sw=4 ts=4:
# SPDX-FileCopyrightText: 2024-present Atri Bhattacharya <atrib@duck.com>
#
# SPDX-License-Identifier: MIT
#

"""
Module that sets up argparse and returns parsed arguments from the cmdline
"""

import argparse
from pathlib import Path
from textwrap import wrap

from rich import print as richprint
from rich.markdown import Markdown
from rich.syntax import Syntax

from tartex.__about__ import __appname__ as APPNAME  # noqa: N812
from tartex.__about__ import __version__
from tartex._completion import (
    COMPFILE,
    BashCompletion,
    FishCompletion,
    ZshCompletion,
)

# Latexmk allowed compilers
LATEXMK_TEX = [
    "dvi",
    "lualatex",
    "luatex",
    "pdf",
    "pdflua",
    "ps",
    "xdv",
    "xelatex",
]


BASH_COMP_PATH = BashCompletion().install_dir.joinpath(f"{APPNAME}")
COMPLETIONS_GUIDE = f"""
Completions are currently supported for bash, fish, and zsh shells.
Please consider [contributing](https://github.com/badshah400/tartex) if you
would like completion for any other shell.

__Note__: XDG_DATA_HOME defaults to `~/.local/share`.

## Bash ##
The option `--bash-completion` will install bash completions for {APPNAME} to
the directory: $XDG_DATA_HOME/{COMPFILE["bash"]}.

Bash automatically searches this dir for completions, so completion for
tartex should work immediately after starting a new terminal session.  If it
does not, you may have to add the following lines to your
'~/.bashrc' file:

```bash
# Source {APPNAME} completion
source ~/{BASH_COMP_PATH.relative_to(Path.home())}
```

## Zsh ##
The option `--zsh-completion` will install a zsh completions file for {APPNAME}
to the directory: $XDG_DATA_HOME/{COMPFILE['zsh'].parent!s}.  It will also
print what to add to your .zshrc file to enable these completions.

## Fish ##
The option `--fish-completion` will install a fish completions file for
{APPNAME} to the directory: $XDG_DATA_HOME/{COMPFILE['fish'].parent!s}.

No further configuration should be needed. Simply start a new fish terminal et
voila!
"""

ZSH_GUIDE = f"""# Update fpath to include completions dir
# Note: Must be done before initialising compinit
fpath=(~/{ZshCompletion().install_dir.relative_to(Path.home())} $fpath)

# If the following two lines already appear in your .zshrc do not add them
# again, but move the fpath line above the 'autoload compinit' line
autoload -U compinit
compinit"""


class CompletionPrintAction(argparse.Action):

    """
    Defines CompletionAction for argparse which will print a completion syntax
    and exit
    """

    def __init__(
        self,
        option_strings,
        dest=argparse.SUPPRESS,
        default=argparse.SUPPRESS,
        help=None,  # noqa: A002
    ):
        """Initialise Action class"""
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help,
        )

    # Note that correct __call__ signature requires all positional args even if
    # they are not used in this method itself
    def __call__(self, parser, nsp, vals, opt_str=None):  # noqa: ARG002
        richprint(Markdown(COMPLETIONS_GUIDE))
        parser.exit()


class CompletionInstall(argparse.Action):

    """
    Defines CompletionAction for argparse which will print a completion syntax
    and exit
    """

    def __init__(
        self,
        option_strings,
        dest=argparse.SUPPRESS,
        default=argparse.SUPPRESS,
        help=None,  # noqa: A002
    ):
        """Initialise Action class"""
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help,
        )

    # Note that correct __call__ signature requires all positional args even if
    # they are not used in this method itself
    def __call__(self, parser, namespace, values, option_string=None):  # noqa
        parser.exit()


class BashCompletionInstall(CompletionInstall):

    """Completion install action for Bash shell"""

    def __call__(self, parser, namespace, values, option_strings=None):
        BashCompletion().install()
        super().__call__(parser, namespace, values, option_strings)
        parser.exit()


class FishCompletionInstall(CompletionInstall):

    """Completion install action for Fish shell"""

    def __call__(self, parser, namespace, values, option_strings=None):
        FishCompletion().install()
        super().__call__(parser, namespace, values, option_strings)
        parser.exit()


class ZshCompletionInstall(CompletionInstall):

    """Completion install action for Zsh shell"""

    def __call__(self, parser, namespace, values, option_strings=None):
        ZshCompletion().install()
        richprint(
            "\n"
            "Add the following to your [bold].zshrc[/] if not already present:"
        )
        richprint(Syntax(ZSH_GUIDE, "zsh"))
        super().__call__(parser, namespace, values, option_strings)
        parser.exit()


class GnuStyleHelpFormatter(argparse.HelpFormatter):

    """
    Format help string in GNU style, i.e.

    * `-s, --long           Help string for long`
      for an action that takes no argument
    * `-s, --long=LONG      Help string for long`
      for an action that requires an argument
    """

    def __init__(self, prog):
        """
        Initialise
        """
        argparse.HelpFormatter.__init__(
            self, prog, max_help_position=30, width=80
        )

    def _format_action_invocation(self, action):
        if not action.option_strings:
            default = self._get_default_metavar_for_positional(action)
            (metavar,) = self._metavar_formatter(action, default)(1)
            return metavar

        parts = []

        # if the Optional doesn't take a value, format is:
        #    -s, --long
        if action.nargs == 0:
            parts.extend(action.option_strings)

        # if the Optional takes a value, format is:
        #    -s, --long=ARGS
        else:
            default = self._get_default_metavar_for_optional(action)
            args_string = self._format_args(action, default)
            for option_string in action.option_strings:
                if len(option_string.lstrip("-")) == 1:  # short form
                    parts.append(f"{option_string}")
                else:  # long form
                    parts.append(f"{option_string}={args_string}")

        return ", ".join(parts)

    def _split_lines(self, text, width):  # noqa: ARG002
        return wrap(text, width=52, break_on_hyphens=False)


def parse_args(args):
    """Set up argparse options and parse input args accordingly"""
    parser = argparse.ArgumentParser(
        description=(
            "Build a tarball including all source files needed to compile your"
            f" LaTeX project (version {__version__})."
        ),
        formatter_class=GnuStyleHelpFormatter,
        usage="%(prog)s [OPTIONS] FILENAME",
    )

    parser.add_argument(
        "fname",
        metavar="FILENAME",
        type=Path,
        help="input file name (with .tex or .fls suffix)",
    )

    parser.add_argument(
        "-V",
        "--version",
        help="print %(prog)s version and exit",
        action="version",
        version=f"%(prog)s {__version__}",
    )

    parser.add_argument(
        "-a",
        "--add",
        metavar="PATTERNS",
        type=str,
        help=(
            "include additional files matching glob-style PATTERN;"
            " separate multiple PATTERNS using commas"
        ),
    )

    parser.add_argument(
        "-b",
        "--bib",
        action="store_true",
        help="find and add bib file to tarball",
    )

    parser.add_argument(
        "-l",
        "--list",
        action="store_true",
        help="print list of files to include and quit",
    )

    parser.add_argument(
        "-o",
        "--output",
        metavar="NAME[.SUF]",
        type=Path,
        help=(
            "output tar file name; tar compression mode will be inferred from"
            " .SUF, if possible (default 'gz')"
        ),
    )

    parser.add_argument(
        "-p",
        "--packages",
        action="store_true",
        help="add names of used (La)TeX packages as a json file",
    )

    parser.add_argument(
        "-s",
        "--summary",
        action="store_true",
        help="print a summary at the end",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        help="increase verbosity (-v, -vv, etc.)",
        action="count",
        default=0,
    )

    parser.add_argument(
        "-x",
        "--excl",
        metavar="PATTERNS",
        type=str,
        help="exclude file names matching PATTERNS",
    )

    # Latexmk options
    latexmk_opts = parser.add_argument_group("Options for latexmk processing")
    latexmk_opts.add_argument(
        "--latexmk-tex",
        metavar="TEXMODE",
        choices=LATEXMK_TEX,
        default=None,
        help=(
            "force TeX processing mode used by latexmk;"
            f" TEXMODE must be one of: {', '.join(LATEXMK_TEX)}"
        ),
    )

    latexmk_opts.add_argument(
        "-F",
        "--force-recompile",
        action="store_true",
        help="force recompilation even if .fls exists",
    )

    # Tar recompress options
    tar_opts = parser.add_mutually_exclusive_group()

    def cmp_str(cmp, ext):
        return f"recompress with {cmp} (.{ext})" " (overrides .SUF in '-o')"

    tar_opts.add_argument(
        "-j",
        "--bzip2",
        action="store_true",
        help=cmp_str("bzip2", "bz2"),
    )

    tar_opts.add_argument(
        "-J",
        "--xz",
        action="store_true",
        help=cmp_str("lzma", "xz"),
    )

    tar_opts.add_argument(
        "-z",
        "--gzip",
        action="store_true",
        help=cmp_str("gzip", "gz"),
    )

    misc_opts = parser.add_argument_group("Shell completion options")
    misc_opts.add_argument(
        "--completion",
        help="print shell completion guides for %(prog)s",
        action=CompletionPrintAction,
    )

    misc_opts.add_argument(
        "--bash-completions",
        help="install bash completions for %(prog)s",
        action=BashCompletionInstall,
    )

    misc_opts.add_argument(
        "--fish-completions",
        help="install fish completions for %(prog)s",
        action=FishCompletionInstall,
    )

    misc_opts.add_argument(
        "--zsh-completions",
        help="install zsh completions for %(prog)s",
        action=ZshCompletionInstall,
    )

    return parser.parse_args(args)



# vim:set et sw=4 ts=4 tw=79:
"""
Tests for package list inclusion
"""

import json
import tarfile as tar

from tartex.tartex import TAR_DEFAULT_COMP, TarTeX


def test_float_pkg(datadir, tmpdir, capsys):
    """
    Test 'packages.json' for float package included from test_packages
    """
    t = TarTeX(
        [
            str(datadir / "main.tex"),
            "-s",
            "-p",
            "-o",
            f"{tmpdir!s}/packagelist",
        ]
    )
    t.tar_files()
    assert "2 files" in capsys.readouterr().out
    with tar.open(
        f"{tmpdir!s}/packagelist.tar.{TAR_DEFAULT_COMP}", mode="r"
    ) as f:
        assert "TeXPackages.json" in f.getnames()

    pkgjson = json.loads(t.pkglist)
    assert "float" in pkgjson["System"]
    assert pkgjson["Local"] == []



# vim:set et sw=4 ts=4:
# SPDX-FileCopyrightText: 2024-present Atri Bhattacharya <atrib@duck.com>
#
# SPDX-License-Identifier: MIT
#

"""
Tests for when source dir has .fls file
"""

import tarfile as tar

import pytest

from tartex.tartex import TAR_DEFAULT_COMP, TarTeX


@pytest.fixture
def flsfile():
    """Default fls file name"""
    return "main.fls"


@pytest.fixture
def tartex_obj(datadir, flsfile):
    """Default tartex object with flsfile as input"""
    return TarTeX(
        [str(datadir / flsfile), "-o", str(datadir), "-bs"],
    )


def test_fls_main_arg(tartex_obj, flsfile):
    """User passes a .fls filename as input"""
    t = tartex_obj
    assert t.tar_file.name == flsfile.replace(".fls", ".tar")


def test_fls_missing_bbl(tartex_obj, flsfile):
    """
    Verify that missing .bbl file is omitted from tarball and the cal to
    tar_files() does not cause an exception
    """
    tartex_obj.tar_files()

    with tar.open(f"{tartex_obj.tar_file!s}.{TAR_DEFAULT_COMP}") as f:
        assert flsfile.replace(".fls", ".bbl") not in f.getnames()


def test_fls_recompile(datadir, flsfile):
    """Forcing a LaTeX recompile ensures .bbl is included"""
    t = TarTeX([str(datadir / flsfile), "-o", str(datadir), "-b", "-F"])
    t.tar_files()

    with tar.open(f"{t.tar_file!s}.{TAR_DEFAULT_COMP}") as f:
        assert flsfile.replace(".fls", ".bbl") in f.getnames()



# vim:set et sw=4 ts=4 tw=80:
# SPDX-FileCopyrightText: 2024-present Atri Bhattacharya <atrib@duck.com>
#
# SPDX-License-Identifier: MIT
#
"""
Test inclusion and exclusion of .bib files
"""

import tarfile as tar

from tartex.tartex import TAR_DEFAULT_COMP, TarTeX


def test_bib(datadir, tmpdir):
    """Test if tar file contains .bib when -b is passed"""
    t = TarTeX(
        [
            str(datadir / "main_bib.tex"),
            "-b",
            "-s",
            "-v",
            "-o",
            f"{tmpdir!s}/main_bib",
        ]
    )
    t.tar_files()
    with tar.open(t.tar_file.with_suffix(f".tar.{TAR_DEFAULT_COMP}")) as f:
        assert "refs.bib" in f.getnames()  # Check: .bib file is in tarball
        # Check main_bib.bbl file is in tarball even though not in srcdir
        assert "main_bib.bbl" in f.getnames()
        # Check user/group name attributes
        for attr in ["gname", "uname"]:
            assert (
                f.getmember("main_bib.bbl").get_info()[attr]
                == f.getmember(t.main_file.name).get_info()[attr]
            )



# vim: set ai et ts=4 sw=4:
# SPDX-FileCopyrightText: 2024-present Atri Bhattacharya <atrib@duck.com>
#
# SPDX-License-Identifier: MIT
#
"""tartex module"""

import fnmatch
import json
import logging as log
import math
import os
import re
import sys
import tarfile as tar
import time
from contextlib import suppress
from io import BytesIO
from pathlib import Path
from tempfile import TemporaryDirectory

from rich import print as richprint
from rich.prompt import Prompt

from tartex import _latex
from tartex._parse_args import parse_args

# Auxilliary file extensions to ignore
# taken from latexmk manual:
# https://www.cantab.net/users/johncollins/latexmk/latexmk-480.txt
AUXFILES = [
    ".aux",
    ".bcf",
    ".fls",
    ".idx",
    ".lof",
    ".lot",
    ".out",
    ".toc",
    ".blg",
    ".ilg",
    ".log",
    ".xdv",
    ".fdb_latexmk",
]

# Supplementary files that are usually required as part of tarball
SUPP_REQ = ["bbl", "ind"]

# Allowed tar extensions
TAR_EXT = ["bz2", "gz", "xz"]

# Default compression
TAR_DEFAULT_COMP = "gz"


def strip_tarext(filename):
    """Strip '.tar(.EXT)' from filename"""
    basename = Path(filename)
    if basename.suffix.lstrip(".") in TAR_EXT:
        basename = basename.with_suffix("")
    if basename.suffix == ".tar":
        basename = basename.with_suffix("")
    return basename


def _full_if_not_rel_path(src, dest):
    p = Path(src).resolve()
    with suppress(ValueError):
        p = p.relative_to(dest)
    if p.is_absolute():
        # Here p is only absolute if it cannot be resolved wrt to dest
        log.debug(
            "Cannot resolve %s relative to %s, will use full path",
            src.as_posix(),
            dest.as_posix(),
        )
    return p


def _summary_msg(nfiles, tarname=None):
    """Return summary msg to print at end of run"""

    def _num_tag(n: int):
        return f"[bold]{n} file" + ("s" if n > 1 else "") + "[/]"

    if tarname:
        richprint(
            f"[cyan]Summary: [bold]{tarname}[/] generated with"
            f" {_num_tag(nfiles)}.[/]"
        )
    else:
        richprint(f"[cyan]Summary: {_num_tag(nfiles)} to include.[/]")


class TarTeX:
    """
    Class to help build  a tarball including all source files needed to
    re-compile your LaTeX project."
    """

    # pylint: disable=too-many-instance-attributes

    pkglist_name = "TeXPackages.json"

    def __init__(self, args):
        self.args = parse_args(args)
        log.basicConfig(
            format="[%(levelname)s] %(message)s",
            level=(
                (log.INFO // self.args.verbose)
                if self.args.verbose > 0
                else log.WARN
            ),
        )
        self.cwd = Path.cwd()
        self.main_file = self.args.fname.resolve()
        if self.main_file.suffix not in [".fls", ".tex"]:
            sys.exit("Error: Source filename must be .tex or .fls\nQuitting")

        # Set default tar extension...
        self.tar_ext = TAR_DEFAULT_COMP
        # ..but use specified output's TAR_EXT extension if any...
        if self.args.output:
            self.args.output = self._proc_output_path()

        # ...but overwrite TAR_EXT if tar compression option passed
        if self.args.bzip2:
            self.tar_ext = "bz2"
        if self.args.gzip:
            self.tar_ext = "gz"
        if self.args.xz:
            self.tar_ext = "xz"

        tar_base = (
            Path(f"{self.args.output}.tar")
            if self.args.output
            else Path(self.main_file.stem).with_suffix(".tar")
        )
        self.tar_file = self.cwd / tar_base  # returns tar_base when absolute
        log.debug(
            "Output tarball '%s' will be generated",
            self.tar_file.with_suffix(f".tar.{self.tar_ext}"),
        )

        self.req_supfiles = {}
        self.add_files = self.args.add.split(",") if self.args.add else []
        excludes = self.args.excl.split(",") if self.args.excl else []
        excl_lists = (self.main_file.parent.glob(f"{L}") for L in excludes)

        self.excl_files = [
            f.relative_to(self.main_file.parent).as_posix()
            for L in excl_lists
            for f in L
        ]

        # If .bbl/.ind is missing in source dir, then it is not in
        # self.excl_files (result of globbing files in srcdir) even if the user
        # passes a matching wildcard to exclude it with "-x".
        # Extend self.excl_files specifically in these cases
        for glb in excludes:
            self.excl_files.extend(
                [
                    f
                    for f in [
                        self.main_file.with_suffix(f".{g}") for g in SUPP_REQ
                    ]
                    if fnmatch.fnmatch(f.name, glb)
                ]
            )

        if self.excl_files:
            log.debug(" ".join(excl_lists))
            log.info(
                "List of excluded files: %s",
                ", ".join([x.as_posix() for x in self.excl_files]),
            )

        self.force_tex = self.args.latexmk_tex
        if not self.force_tex:
            # If force_tex is not set by user options,
            # set to ps if source dir contains ps/eps files
            # or to pdf otherwise
            src_ps = [
                str(p)
                for ext in ["eps", "ps"]
                for p in self.main_file.parent.glob(f"**/*.{ext}")
            ]
            self.force_tex = "ps" if src_ps else "pdf"
            log.info(
                "Latexmk will use %slatex for processing, if needed",
                self.force_tex,
            )

        self.pkglist = None

    def add_user_files(self):
        """
        Return list of additional user specified/globbed file paths,
        if they exist
        """
        lof = []
        for fpatt in self.add_files:
            afiles = list(self.main_file.parent.glob(fpatt))
            if not afiles:
                log.warning(
                    "No match corresponding to user specified pattern '%s' for"
                    " additional files",
                    fpatt,
                )
                continue
            lof.extend(afiles)

        return lof

    def bib_file(self):
        """Return relative path to bib file"""
        bibre = re.compile(r"^\\bibliography\{.*\}")
        bibstr = None
        texf = self.main_file.with_suffix(".tex")
        with open(texf, encoding="utf-8") as f:
            for line in f:
                m = bibre.search(line)
                if m:
                    bibstr = m.group()
                    break

        if bibstr:
            bibstr = re.sub(r"^\\bibliography\{", "", bibstr).rstrip("}")
            bibstr += ".bib" if bibstr.split(".")[-1] != ".bib" else ""

        return Path(bibstr) if bibstr else None

    def input_files(self):
        """
        Returns non-system input files needed to compile the main tex file.
        Will try to compile the main tex file using `latexmk` if it cannot find
        the fls file in the same dir.
        """
        if (
            not self.main_file.with_suffix(".fls").exists()
            or self.args.force_recompile
        ):
            with TemporaryDirectory() as compile_dir:
                log.info(
                    "LaTeX recompile forced"
                    if self.args.force_recompile
                    else f"{self.main_file.stem}.fls file not found in"
                    f" {self.main_file.parent.as_posix()}"
                )
                log.info("LaTeX compile directory: %s", compile_dir)
                fls_path = _latex.run_latexmk(
                    self.main_file.with_suffix(".tex"),
                    self.force_tex,
                    compile_dir,
                )

                with open(fls_path, encoding="utf-8") as f:
                    deps, pkgs = _latex.fls_input_files(
                        f,
                        self.excl_files,
                        AUXFILES,
                        sty_files=self.args.packages,
                    )
                    if self.args.packages:
                        log.info(
                            "System TeX/LaTeX packages used: %s",
                            ", ".join(sorted(pkgs["System"]))
                        )

                        self.pkglist = json.dumps(pkgs, cls=SetEncoder).encode(
                            "utf8"
                        )

                for ext in SUPP_REQ:
                    if app := self._missing_supp(
                        self.main_file.with_suffix(f".{ext}"), compile_dir, deps
                    ):
                        self.req_supfiles[
                            self.main_file.with_suffix(f".{ext}")
                        ] = app
        else:
            # If .fls exists, this assumes that all INPUT files recorded in it
            # are also included in source dir
            with open(self.main_file.with_suffix(".fls"), encoding="utf8") as f:
                deps, pkgs = _latex.fls_input_files(
                    f, self.excl_files, AUXFILES, sty_files=self.args.packages
                )
                if self.args.packages:
                    self.pkglist = json.dumps(pkgs, cls=SetEncoder).encode(
                        "utf8"
                    )

        if self.args.bib and (bib := self.bib_file()):
            deps.append(bib.as_posix())
            log.info("Add file: %s", deps[-1])

        if self.add_files:
            for f in self.add_user_files():
                f_relpath_str = f.relative_to(self.main_file.parent).as_posix()
                if f_relpath_str in deps:
                    log.warning(
                        "Manually included file %s already added", f_relpath_str
                    )
                    continue
                deps.append(f_relpath_str)
                log.info("Add user specified file: %s", f_relpath_str)

        return deps

    def tar_files(self):
        """
        Generates a tarball consisting of non-system input files needed to
        recompile your latex project.
        """
        self.check_main_file_exists()
        full_tar_name = Path(f"{self.tar_file}.{self.tar_ext}")

        wdir = self.main_file.resolve().parent
        os.chdir(wdir)
        log.debug("Switching working dir to %s", wdir.as_posix())
        if self.args.list:
            self._print_list(self.input_files())
        else:
            try:
                f = tar.open(full_tar_name, mode=f"x:{self.tar_ext}")
                with f:
                    self._do_tar(f)
                    if self.args.summary:
                        _summary_msg(
                            len(f.getmembers()),
                            _full_if_not_rel_path(full_tar_name, self.cwd),
                        )
            except PermissionError as err:
                log.critical(
                    "Cannot write to %s, %s",
                    full_tar_name.parent,
                    err.strerror.lower(),
                )
                sys.exit(1)
            except FileExistsError:
                try:
                    full_tar_name = self._tar_name_conflict(full_tar_name)
                    # At this stage, there is either a new name for the tar
                    # file or user wants to overwrite existing file. In either
                    # case, calling tar.open() with 'w' mode should be OK.
                    f = tar.open(full_tar_name, mode=f"w:{self.tar_ext}")
                    with f:
                        self._do_tar(f)
                        if self.args.summary:
                            _summary_msg(
                                len(f.getmembers()),
                                _full_if_not_rel_path(full_tar_name, self.cwd),
                            )
                except PermissionError as err:
                    log.critical(
                        "Cannot write to %s, %s",
                        full_tar_name.parent,
                        err.strerror.lower(),
                    )
                    sys.exit(1)
        os.chdir(self.cwd)
        log.debug("Reset working dir to %s", os.getcwd())

    def _tar_name_conflict(self, tpath):
        richprint(
            "[bold red]A tar file with the same name"
            rf" \[{_full_if_not_rel_path(tpath, self.cwd)}]"
            " already exists[/bold red]"
        )

        owr = Prompt.ask(
            "What would you like to do "
            r"([bold blue]\[O][/bold blue]verwrite /"
            r" [bold green]\[C][/bold green]hoose new name /"
            r" *[bold red]\[Q][/bold red]uit)?"
        )
        if owr.lower() in ["", "q"]:
            richprint(
                "[bold]Not overwriting existing tar file[/bold]\nQuitting",
                file=sys.stderr,
            )
            sys.exit(1)
        elif owr.lower() == "c":
            new_name = Prompt.ask("Enter [bold]new name[/bold] for tar file")
            if (new_ext := new_name.split(".")[-1]) in TAR_EXT:
                self.tar_ext = new_ext
            # If new file is a plain file name, interpret w.r.t. output dir
            if self.args.output and (
                str(Path(new_name)) == Path(new_name).name
            ):
                new_name = Path(self.args.output).with_name(new_name)
            else:
                new_name = self.cwd / Path(new_name).expanduser()
            new_path = new_name.with_name(
                f"{strip_tarext(new_name.name)}.tar.{self.tar_ext}"
            ).resolve()
            if new_path == tpath:
                richprint(
                    "[bold red]Error: New name entered is also the same as"
                    " existing tar file name[/bold red]\nQuitting",
                    file=sys.stderr,
                )
                sys.exit(1)
            elif new_path.exists():
                richprint(
                    "[bold red]Error: A tar file with the same name"
                    rf" \[{_full_if_not_rel_path(new_path, self.cwd)!s}] also"
                    " exists[/bold red]\nQuitting",
                    file=sys.stderr,
                )
                sys.exit(1)
            else:
                log.info("Tar file %s will be generated", new_path.as_posix())
                return new_path
        elif owr.lower() == "o":
            return tpath
        else:
            richprint(
                "[bold red]Error: Invalid response[/]\nQuitting.",
                file=sys.stderr,
            )
            sys.exit(1)

    def check_main_file_exists(self):
        """Check for the existence of the main tex/fls file."""
        if not Path(self.main_file).exists():
            log.critical("File not found - %s", self.main_file)
            sys.exit(1)

    def _do_tar(self, tar_obj):
        for dep in self.input_files():
            # By now, if the file is still missing, this indicates a .fls file
            # is present in source but some of its INPUT marked entries are
            # not. That is user error and we simply omit the missing file from
            # tarball with a warning
            try:
                tar_obj.add(dep)
            except FileNotFoundError:
                log.warning(
                    "Skipping INPUT file '%s', not found amongst sources; try"
                    " forcing a LaTeX recompile ('-F').",
                    dep,
                )

        def _tar_add_bytesio(obj, name):
            tinfo = tar_obj.tarinfo(name)
            tinfo.size = len(obj)
            tinfo.mtime = int(time.time())

            # Copy user/group names from main.tex file
            tinfo.uname = tar_obj.getmember(
                self.main_file.with_suffix(".tex").name
            ).uname
            tinfo.gname = tar_obj.getmember(
                self.main_file.with_suffix(".tex").name
            ).gname
            tar_obj.addfile(tinfo, BytesIO(obj))

        if self.args.packages:
            log.info(
                "Adding list of packages as BytesIO object: %s",
                self.pkglist_name,
            )
            _tar_add_bytesio(self.pkglist, self.pkglist_name)

        for fpath, byt in self.req_supfiles.items():
            log.info("Adding %s as BytesIO object", fpath.name)
            _tar_add_bytesio(byt, fpath.name)

    def _proc_output_path(self):
        """
        Returns the output tar file path (sans any '.tar.?z' suffix) after
        resolving the value passed to '--output' as part of cmdline options.

        Also sets the tar ext if determined from '--output' argument.
        """

        # If self.args.output is absolute, '/' simply returns it as a PosixPath
        out = self.cwd / self.args.output.expanduser()

        if out.is_dir():  # If dir, set to DIR/main.tar.gz
            log.debug("%s is an existing dir", out)
            out = out.joinpath(
                self.main_file.with_suffix(f".tar.{TAR_DEFAULT_COMP}").name
            )
        elif (ext := out.suffix.lstrip(".")) in TAR_EXT:
            self.tar_ext = ext
        else:
            out = out.with_name(f"{out.name}.tar.{TAR_DEFAULT_COMP}")

        out = strip_tarext(out)
        log.debug("Processed output target: %s", self.args.output)
        return out.as_posix()

    def _print_list(self, ls):
        """helper function to print list of files in a pretty format"""
        idx_width = int(math.log10(len(ls))) + 1
        for i, f in enumerate(ls):
            richprint(f"{i+1:{idx_width}}. {f}")
        for r in self.req_supfiles:
            richprint(f"{'*':>{idx_width + 1}} {r.name}")
        if self.args.packages:
            richprint(f"{'*':>{idx_width + 1}} {self.pkglist_name}")
        if self.args.summary:
            _summary_msg(
                len(ls) + len(self.req_supfiles) + (1 if self.pkglist else 0)
            )

    def _missing_supp(self, fpath, tmpdir, deps):
        """Handle missing supplementary file from orig dir, if req"""
        if (
            fpath not in deps  # bbl file not in source dir
            and (Path(tmpdir) / fpath.name).exists()  # Implies it's req
            and fpath not in self.excl_files  # Not explicitly excluded
        ):
            log.debug("Required file '%s' not in source dir", fpath.name)
            log.info(
                "Add contents as BytesIO: %s",
                Path(tmpdir) / fpath.name,
            )
            return (Path(tmpdir) / fpath.name).read_bytes()

        return None


def make_tar():
    """Build tarball with command line arguments processed by argparse"""
    t = TarTeX(sys.argv[1:])
    t.tar_files()


class SetEncoder(json.JSONEncoder):
    """A class to allow JSONEncoder to interpret a set as a list"""

    def default(self, o):
        """
        Convert o (a set) into a sorted list

        :o: set
        :return: list
        """
        return sorted(o)


if __name__ == "__main__":
    make_tar()
